# Last update on 20 Nov, 2017
# this program analyzes the TCGA data using LASSO, elastic net, or I-Boost
# the program file "boosting_functions.R", data file "TCGA_8cancer_rmmis.csv",
#	and the file "DataSplit.csv" indicating how subjects should be split
#	should be present in the home directory



### Analysis Settings:
method <- "CV" ### method="permute" for I-Boost-Permutation;
			### method="CV" for I-Boost-CV;
			### method="lasso" for LASSO;
			### method="en" for elastic net;
type <- 1 ### subset of subjects to be analyzed (1=LUAD, 2=KIRC, 3=pan-cancer)
predset <- 81 ### the set of predictors; see the file index.csv to see the label for
		### each combination of data type
s <- 1 ### the column number in the file "DataSplit.csv" to use


library(Hmisc)
library(glmnet)
library(R.utils)
library(survival)
library(data.table)
source("boosting_functions.R")

for (j in 1:6) {
	if (j == 1) rawdata <- data.frame(fread("TCGA_8cancer_rmmis_1.csv",header=TRUE,sep=","),row.names=1) else
		rawdata <- rbind(rawdata,data.frame(fread(paste("TCGA_8cancer_rmmis_",j,".csv",sep=""),header=TRUE,sep=","),row.names=1))
}
x <- list(outcome=t(rawdata[1:2,]),xd=rawdata[-c(1:2),],fnames=rownames(rawdata)[-c(1,2)],snames=colnames(rawdata))

Clinical <- grep("Clinical_",x$fnames)
Clinical <- setdiff(Clinical, which(x$fnames=="Clinical_BRCA"))
CNV <- grep("CNV_",x$fnames)
miRNA <- grep("miRNA_",x$fnames)
Mutation <- grep("Mutation_",x$fnames)
Module <- grep("Module_",x$fnames)
Protein <- grep("Protein_",x$fnames)
GeneExp <- grep("Gene_",x$fnames)
data.types <- list(Clinical=Clinical,GeneExp=GeneExp,Module=Module,Protein=Protein,miRNA=miRNA,CNV=CNV,Mutation=Mutation)
data.types.names <- list("Clinical","Gene","Module","Protein","miRNA","CNV","Mutation")

features.rows <- list(Clinical,CNV,Mutation,miRNA,Protein)
features <- c("Clinical","CNV","Mutation","miRNA","Protein")
m <- length(features)

FEATURES <- list()
NAMES <- rep(NA, 2^m)
for (i in 0:(2^m-1)) {
	a <- unlist(strsplit(intToBin(i), split=""))
	a <- c(rep(0,m-length(a)), a)
	a <- rev(a)
	index <- c("")
	name <- c("")
	feature.tmp <- c()
	for (j in 1:m) {
		if (a[j] == "1") {
			index <- paste(index,j+1,sep="")
			name <- paste(name,features[j],sep=".")
			feature.tmp <- c(feature.tmp, features.rows[[j]])
		}
	}
	NAMES[i+1] <- paste(index,name,sep="")
	FEATURES <- c(FEATURES, list(feature.tmp))
}
NAMES[1] <- c(".")

insert.string <- function(name, index, feature) {
	a <- unlist(strsplit(name, split="\\."))
	if (length(a) == 0) a <- c("")
	out <- paste(c(paste(index,a[1],sep=""),feature,a[-1]), collapse=".")
	c(out)
}
combine <- function(x, object) c(object, x)

NAMES.95 <- as.vector(c(sapply(NAMES, insert.string, index=0, feature="GeneExp"), sapply(NAMES, insert.string, index=1, feature="Module"), NAMES[-1]))
FEATURES.95 <- c(lapply(FEATURES, combine, object=GeneExp), lapply(FEATURES, combine, object=Module), FEATURES[-1])
n.model <- length(NAMES.95)

KMplot <- function(x, event, stime, ylab="Overall Survival (Probability)", xlab="Years", varName) {
	mainLabel <- varName
	fit <- survfit(Surv(stime, event) ~ x)
	
	plot(fit, col=2:3, main=varName, ylim=c(0,1), ylab=ylab, xlab=xlab)
}
cox.pvalue <- function(X,Y,Z) {
	fit <- coxph(Y ~ cbind(X,Z), ties="breslow")
	c(ifelse(is.na(summary(fit)$coefficients[1,5]), 1, summary(fit)$coefficients[1,5]))
}

name <- NAMES.95[predset]
features <- FEATURES.95[[predset]]

type.name <- ifelse(type == 1, "LUAD", ifelse(type == 2, "KIRC", "all"))
out <- c()
out <- c(out, name)

if (type == 1) filter.t <- (x$xd["Clinical_LUAD",] == 1)
if (type == 2) filter.t <- (x$xd["Clinical_KIRC",] == 1)
if (type == 3) filter.t <- rep(TRUE,ncol(x$xd))
type.name <- ifelse(type == 1, "LUAD", ifelse(type == 2, "KIRC", "all"))
rm.features <- c()
for (k in Clinical) if (length(unique(as.numeric(x$xd[k,filter.t]))) == 1) rm.features <- c(rm.features, k)
features.t <- setdiff(features, rm.features)
all.modules <- t(x$xd[features.t, filter.t])
all.outcome <- Surv(x$outcome[filter.t,2],x$outcome[filter.t,1])

split <- read.table("DataSplit.csv",sep=",",header=TRUE,row.names=1)
split <- split[match(x$snames,rownames(split)),]

train.modules <- t(x$xd[features.t, filter.t & split[,s] == 0])
train.outcome <- Surv(x$outcome[filter.t&split[,s]==0,2],x$outcome[filter.t&split[,s]==0,1])
test.modules <- t(x$xd[features.t, filter.t & split[,s] == 1])
test.outcome <- Surv(x$outcome[filter.t&split[,s]==1,2],x$outcome[filter.t&split[,s]==1,1])

train.modules <- t(t(train.modules) - apply(train.modules,2,mean))
train.modules.sd <- apply(train.modules,2,function(x)ifelse(sd(x)>0,sd(x),1))
train.modules <- t(t(train.modules) / train.modules.sd)


procedure <- ifelse(method=="permute"|method=="CV","IB","EN")


if (procedure == "EN") {
	n.repeat.cv <- 5
        if (method == "lasso") alpha <- c(1) else alpha <- c(0.05, seq(0.1,1,by=0.1))
        foldids <- list()
        for (j in 1:n.repeat.cv) {
                set.seed(i * 31415 + 1938274 * j)
                foldid <- rep(cumsum(rep(1,5)), length=nrow(train.modules))
                foldid <- sample(foldid, replace=F)
                foldids <- c(foldids, list(foldid))
        }
        min.cvm <- 1e8
        out.cvm <- c()
        for (a in alpha) {
                cvm <- rep(0, 100)
                for (j in 1:n.repeat.cv) {
                        cv.fit.en <- manual.cv.glmnet(train.modules, train.outcome, family="cox", alpha=a, foldid=foldids[[j]], error="cox", naive=TRUE)
                        short <- min(length(cvm),length(cv.fit.en$lambda))
                        cvm <- cvm[1:short] + cv.fit.en$cvm[1:short] / n.repeat.cv
                }
                out.cvm <- c(out.cvm, min(cvm))
                if (min(cvm) < min.cvm) {
                        best.alpha = a
                        best.lambda = cv.fit.en$lambda[1:which.min(cvm)]
                        min.cvm <- min(cvm)
                }
        }
        fit.en <- glmnet(train.modules, train.outcome, family="cox", alpha=best.alpha, lambda=best.lambda)
	beta <- fit.en$beta[,length(best.lambda)]
	tuning <- c(best.alpha,tail(best.lambda,1))
} else if (procedure == "IB") {
	sub.data.types <- lapply(data.types,function(x)match(intersect(x,features.t),features.t))
	sub.data.types <- sub.data.types[as.numeric(lapply(sub.data.types,function(x)ifelse(length(x)==0,0,1)))==1]
	alpha.series <- c(0.05,seq(0.1,1,by=0.1))
	if (method == "permute") v <- 0.1 else v <- 1
	ib <- IBoost(train.modules,train.outcome,sub.data.types,alpha.series=alpha.series,v=v,method=method,seed=123987*i,iter.max=2000)
	beta <- ib$beta
	it <- ib$iter.no
	tuning <- c(it,NA)
}
out <- c(out, tuning)
beta <- beta / train.modules.sd
train.modules <- t(t(train.modules) * train.modules.sd)
selected <- x$fnames[features.t[which(beta != 0)]]
selected.pos <- which(beta != 0)
beta.nonzero <- beta[which(beta != 0)]
out <- c(out, length(beta.nonzero))

if (length(grep("GeneExp", name)) > 0) GeneExpsize <- length(grep("Gene_", selected)) else GeneExpsize <- c(NA)
if (length(grep("Module", name)) > 0) Modulesize <- length(grep("Module_", selected)) else Modulesize <- c(NA)
if (length(grep("Clinical", name)) > 0) Clinicalsize <- length(grep("Clinical_", selected)) else Clinicalsize <- c(NA)
if (length(grep("CNV", name)) > 0) DNAsize <- length(grep("CNV_", selected)) else DNAsize <- c(NA)
if (length(grep("Mutation", name)) > 0) Mutationsize <- length(grep("Mutation_", selected)) else Mutationsize <- c(NA)
if (length(grep("miRNA", name)) > 0) miRNAsize <- length(grep("miRNA_", selected)) else miRNAsize <- c(NA)
if (length(grep("Protein", name)) > 0) RPPAsize <- length(grep("Protein_", selected)) else RPPAsize <- c(NA)
out <- c(out, GeneExpsize, Modulesize, Clinicalsize, DNAsize, Mutationsize, miRNAsize, RPPAsize)

outmodel <- cbind(selected[order(beta.nonzero)], beta.nonzero[order(beta.nonzero)])

par(oma=c(0,0,0.5,0))
par(mfrow=c(1,2))

preds <- as.matrix(train.modules[,selected.pos]) %*% beta.nonzero
if (length(unique(preds)) > 1) {
	preds <- (preds - mean(preds)) / sd(preds)
	vals.en <- cut(preds,c(-1000,median(preds),1000),c("low","high"))
	if (length(unique(vals.en)) > 1) plot.KM <- TRUE else plot.KM <- FALSE
	if (plot.KM) {
		log.rank.p <- 1-pchisq(survdiff(Surv(train.outcome[,1], train.outcome[,2]) ~ vals.en)$chisq, 1)
		KMplot(vals.en, train.outcome[,2], train.outcome[,1], varName=paste("Training Set", name), xlab="Month")
	} else log.rank.p <- c(1)
	cindex <- signif(1-rcorr.cens(preds,train.outcome)[[1]],4)
	cox.fit <-summary(coxph(train.outcome~preds, ties="breslow"))
	if (plot.KM) {
		if (cox.fit$coef[,5]<0.05) {cox.p.color<-"darkred"}else{cox.p.color<-"black"}
		legend(3,0.4,legend=paste("C-index =",cindex,  "\nCox HR =", signif(cox.fit$coef[,2],4), "\nCox p =", signif(cox.fit$coef[,5],3),
				"\nlog-rank p =", signif(log.rank.p,3)),text.col=cox.p.color,bty="n")
		risk.event <- table(vals.en,train.outcome[,2])
		legend(3,0.15,legend=c(paste("Low Risk ",risk.event[1,2],"/",sum(risk.event[1,]),sep=""), 
				paste("High Risk ",risk.event[2,2],"/",sum(risk.event[2,]),sep="")) ,text.col=2:3,bty="n")
	}
	out <- c(out, dim(train.modules)[1], log.rank.p[1], signif(cox.fit$coef[,2],4), signif(cox.fit$coef[,5],3), cindex)
} else {
	out <- c(out, rep("NA",5))
}

preds <- as.matrix(test.modules[,selected.pos]) %*% beta.nonzero
if (length(unique(preds)) > 1) {
	preds <- (preds - mean(preds)) / sd(preds)
	vals.en <- cut(preds,c(-1000,median(preds),1000),c("low","high"))
	if (length(unique(vals.en)) > 1) plot.KM <- TRUE else plot.KM <- FALSE
	if (plot.KM) {
		log.rank.p <- 1-pchisq(survdiff(Surv(test.outcome[,1], test.outcome[,2]) ~ vals.en)$chisq, 1)
		KMplot(vals.en, test.outcome[,2], test.outcome[,1], varName=paste("Testing Set", name), xlab="Month")
	} else log.rank.p <- c(1)
	cindex <- signif(1-rcorr.cens(preds,test.outcome)[[1]],4)
	cox.fit<-summary(coxph(test.outcome~preds, ties="breslow"))
	if (plot.KM) {
		if (cox.fit$coef[,5]<0.05) {cox.p.color<-"darkred"}else{cox.p.color<-"black"}
		legend(3,0.4,legend=paste("C-index =",cindex,  "\nCox HR =", signif(cox.fit$coef[,2],4), "\nCox p =", signif(cox.fit$coef[,5],3),
				"\nlog-rank p =", signif(log.rank.p,3)),text.col=cox.p.color,bty="n")
		risk.event <- table(vals.en,test.outcome[,2])
		legend(3,0.15,legend=c(paste("Low Risk ",risk.event[1,2],"/",sum(risk.event[1,]),sep=""), 
				paste("High Risk ",risk.event[2,2],"/",sum(risk.event[2,]),sep="")) ,text.col=2:3,bty="n")
	}
	out <- c(out, dim(test.modules)[1], log.rank.p[1], signif(cox.fit$coef[,2],4), signif(cox.fit$coef[,5],3), cindex)
} else {
	out <- c(out, rep("NA",5))
}

preds <- rep(NA, length=length(x$snames))
preds[filter.t] <- as.matrix(all.modules[,selected.pos]) %*% beta.nonzero
preds[filter.t] <- (preds[filter.t] - mean(preds[filter.t])) / sd(preds[filter.t])
predout <- cbind(as.matrix(x$snames), split[,s], as.matrix(preds))


names(out) <- c("Model","alpha/Iteration No","lambda","Total Size","GeneExp Size","Module Size","Clinical Size","CNV Size",
		"Mutation Size","miRNA Size","Protein Size","Train N","Train Log Rank P","Train HR","Train COX P","Train C-Index",
		"Test N","Test Log Rank P","Test HR","Test COX P","Test C-Index")
colnames(outmodel) <- c("Feature","Coefficient")

outmodel[,"Coefficient"] <- round(as.numeric(outmodel[,"Coefficient"]),4)
colnames(predout) <- c("ID","Testing Set","Risk")
predout <- predout[!is.na(predout[,"Risk"]),]
predout[,"Risk"] <- round(as.numeric(predout[,"Risk"]),4)

### Output:
print(out) # result summary
print(outmodel) # selected model and parameter estimates
print(predout) # predicted risk score

filename <- paste(name,type.name,s,method,sep="-")
write.table(file=paste("summary-",filename,".csv",sep=""),out,sep=",",quote=FALSE,col.names=FALSE)
write.table(file=paste("model-",filename,".csv",sep=""),outmodel,sep=",",quote=FALSE,row.names=FALSE)
write.table(file=paste("predict-",filename,".csv",sep=""),predout,sep=",",quote=FALSE,row.names=FALSE)